import configparser
import json
import logging
from collections import OrderedDict
from pathlib import Path
from typing import Dict, Any, Union

from transfer_nlp.plugins.config import ExperimentConfig
from transfer_nlp.plugins.reporters import ReporterABC
from transfer_nlp.plugins.trainers import TrainerABC

ConfigEnv = Dict[str, Any]

def load_config(p: Path) -> Dict[str, ConfigEnv]:

    def get_val(cfg:configparser.ConfigParser, section: str, key):
        try:
            return cfg.getint(section, key)
        except ValueError:
            pass
        try:
            return cfg.getfloat(section, key)
        except ValueError:
            pass
        try:
            return cfg.getboolean(section, key)
        except ValueError:
            pass

        return cfg[section][key]

    cfg = configparser.ConfigParser()
    cfg.optionxform = str
    cfg.read(p)

    rv = {}

    for exp_name in cfg.sections():
        exp = {}
        for key in cfg[exp_name].keys():
            exp[key] = get_val(cfg, exp_name, key)
        rv[exp_name] = exp

    return rv

class ExperimentRunner:
    """
    Run an experiment several times with varying configurations.

    This class facilitates reusing a single json experiment file across several different configuations.
    """

    @staticmethod
    def _capture_logs(report_path:Path):
        logger = logging.getLogger('')
        handler = logging.FileHandler(str(report_path / 'runner.log'))
        fmt = logging.Formatter('%(asctime)s %(levelname)s: %(message)s')  # TODO configurable?
        handler.setFormatter(fmt)
        logger.addHandler(handler)
        return handler

    @staticmethod
    def _stop_log_capture(handler):
        logger = logging.getLogger('')
        logger.removeHandler(handler)

    @staticmethod
    def _write_config(cfg_name: str, experiment:Dict, cfg:ConfigEnv, exp_report_path:Path):
        """duplicate the config used to run the experiment in the report directory to preserve history"""
        config = configparser.ConfigParser({}, OrderedDict)
        config.optionxform = str
        config.add_section(cfg_name)
        for k in sorted(cfg.keys()):
            config.set(cfg_name, k, str(cfg[k]))

        with (exp_report_path / 'experiment.cfg').open('w') as configfile:
            config.write(configfile)

        with (exp_report_path / 'experiment.json').open('w') as expfile:
            json.dump(experiment, expfile, indent=4)

    @staticmethod
    def run_all(experiment: Union[str, Path, Dict],
                experiment_config: Union[str, Path],
                report_dir: Union[str, Path],
                trainer_config_name: str = 'trainer',
                reporter_config_name: str = 'reporter',
                **env_vars) -> None:
        """
        :param experiment: the experiment config
        :param experiment_config: the experiment config file. The cfg file should be defined in `ConfigParser
               <https://docs.python.org/3/library/configparser.html#module-configparser>`_ format such that
               each section is an experiment configuration.
        :param report_dir: the directory in which to produce the reports. It's recommended to include a timestamp your report directory so you
               can preserve previous reports across code changes. E.g. $HOME/reports/run_2019_02_22.
        :param trainer_config_name: the name of the trainer configuration object. The referenced object should implement `TrainerABC`.
        :param reporter_config_name: the name of the reporter configuration object. The referenced object should implement `ReporterABC`.
        :param env_vars: any additional environment variables, like file system paths
        :return: None
        """

        envs: Dict[str, ConfigEnv] = load_config(Path(experiment_config))

        report_path = Path(report_dir)
        report_path.mkdir(parents=True)

        for exp_name, env in envs.items():
            exp_report_path = report_path / exp_name
            exp_report_path.mkdir()
            log_handler = ExperimentRunner._capture_logs(exp_report_path)
            try:
                logging.info('running %s', exp_name)
                all_vars = dict(env_vars)
                all_vars.update(env)
                experiment_config = ExperimentConfig(experiment, **all_vars)
                trainer: TrainerABC = experiment_config[trainer_config_name]
                reporter: ReporterABC = experiment_config[reporter_config_name]
                trainer.train()
                exp_json = ExperimentConfig.load_experiment_json(experiment)
                ExperimentRunner._write_config(exp_name, exp_json, all_vars, exp_report_path)
                reporter.report(exp_name, experiment_config, exp_report_path)
            finally:
                ExperimentRunner._stop_log_capture(log_handler)



"""
This class aims at using a pytorch loss function into ignite Metrics
"""

import logging

from ignite.metrics import Loss

from transfer_nlp.plugins.config import register_plugin

logger = logging.getLogger(__name__)


@register_plugin
class LossMetric(Loss):
    """
    avoid name collision on batch size param of super class
    """

    def __init__(self, loss_fn):
        super().__init__(loss_fn)



from typing import Dict


class Vocabulary:

    def __init__(self, token2id: Dict = None, add_unk: bool = True, unk_token: str = "<UNK>"):

        if token2id is None:
            token2id = {}

        self._token2id: Dict = token2id
        self._id2token = {idx: token for token, idx in self._token2id.items()}

        self._add_unk: bool = add_unk
        self._unk_token: str = unk_token
        self.unk_index: int = -1
        if add_unk:
            self.unk_index = self.add_token(unk_token)

    @classmethod
    def from_serializable(cls, contents):

        return cls(**contents)

    def to_serializable(self):

        return {
            'token2id': self._token2id,
            'add_unk': self._add_unk,
            'unk_token': self._unk_token}

    def add_token(self, token: str):

        if token in self._token2id:
            index = self._token2id[token]
        else:
            index = len(self._token2id)
            self._token2id[token] = index
            self._id2token[index] = token
        return index

    def add_many(self, tokens):

        return [self.add_token(token) for token in tokens]

    def lookup_token(self, token: str):

        if self._add_unk:
            return self._token2id.get(token, self.unk_index)
        else:
            return self._token2id.get(token, None)

    def lookup_index(self, index: int):

        if index not in self._id2token:
            raise ValueError(f"Index {index} is not present in the Vocabulary")

        else:
            return self._id2token[index]

    def __str__(self):
        return f"Vocabulary(size={len(self)})"

    def __len__(self):
        return len(self._token2id)


class CBOWVocabulary(Vocabulary):

    def __init__(self, token2id: Dict = None, add_unk: bool = True, unk_token: str = "<UNK>", mask_token: str = "<MASK>"):
        super().__init__(token2id=token2id, add_unk=add_unk, unk_token=unk_token)
        self._mask_token = mask_token
        self.mask_index = self.add_token(self._mask_token)

    def to_serializable(self):
        contents = super(CBOWVocabulary, self).to_serializable()
        contents.update({
            'mask_token': self._mask_token})
        return contents


class SequenceVocabulary(Vocabulary):

    def __init__(self, token2id: Dict = None, unk_token: str = "<UNK>",
                 mask_token: str = "<MASK>", begin_seq_token: str = "<BEGIN>",
                 end_seq_token: str = "<END>"):

        super(SequenceVocabulary, self).__init__(token2id=token2id, add_unk=True, unk_token=unk_token)

        self._mask_token: str = mask_token
        self._begin_seq_token: str = begin_seq_token
        self._end_seq_token: str = end_seq_token

        self.mask_index: int = self.add_token(self._mask_token)
        self.begin_seq_index: int = self.add_token(self._begin_seq_token)
        self.end_seq_index: int = self.add_token(self._end_seq_token)

    def to_serializable(self):

        contents = super(SequenceVocabulary, self).to_serializable()

        contents.update({
            'mask_token': self._mask_token,
            'begin_seq_token': self._begin_seq_token,
            'end_seq_token': self._end_seq_token})
        del contents['add_unk']
        return contents

    @classmethod
    def from_serializable(cls, contents):

        return cls(**contents)

    def lookup_token(self, token):

        if self.unk_index >= 0:
            return self._token2id.get(token, self.unk_index)
        else:
            return self._token2id[token]



"""
This class contains the abstraction interface to customize runners.
For the training loop, we use the engine logic from pytorch-ignite

Check experiments for examples of experiment json files

"""
import inspect
import logging
from abc import ABC, abstractmethod
from itertools import zip_longest
from typing import Dict, List, Any
from collections import defaultdict

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from ignite.contrib.handlers.tqdm_logger import ProgressBar
from ignite.engine import Events
from ignite.engine.engine import Engine
from ignite.metrics import Loss, Metric, RunningAverage
from ignite.utils import convert_tensor
from ignite.contrib.handlers.tensorboard_logger import TensorboardLogger, OutputHandler, OptimizerParamsHandler, WeightsScalarHandler, WeightsHistHandler, \
    GradsScalarHandler
from tensorboardX import SummaryWriter


from transfer_nlp.loaders.loaders import DatasetSplits
from transfer_nlp.plugins.config import register_plugin, ExperimentConfig, PluginFactory
from transfer_nlp.plugins.regularizers import RegularizerABC

logger = logging.getLogger(__name__)


def set_seed_everywhere(seed: int, cuda: bool):
    np.random.seed(seed)
    torch.manual_seed(seed)
    if cuda:
        torch.cuda.manual_seed_all(seed)


def _prepare_batch(batch: Dict, device=None, non_blocking: bool = False):
    """Prepare batch for training: pass to a device with options.

    """
    result = {key: convert_tensor(value, device=device, non_blocking=non_blocking) for key, value in batch.items()}
    return result


class TrainingMetric(Metric):

    def __init__(self, metric: Metric):
        self.source_metric = metric
        self.reset()

        super().__init__(lambda x: x[:-1])

    def reset(self):
        self.source_metric.reset()

    def update(self, output):
        self.source_metric.update(output)

    def compute(self):
        return self.source_metric.compute()


class TrainerABC(ABC):

    @abstractmethod
    def train(self):
        pass

@register_plugin
class BasicTrainer(TrainerABC):

    def __init__(self,
                 model: nn.Module,
                 dataset_splits: DatasetSplits,
                 loss: nn.Module,
                 optimizer: optim.Optimizer,
                 metrics: Dict[str, Metric],
                 experiment_config: ExperimentConfig,
                 device: str = None,
                 num_epochs: int = 1,
                 seed: int = None,
                 cuda: bool = None,
                 loss_accumulation_steps: int = 4,
                 scheduler: Any = None,  # no common parent class?
                 regularizer: RegularizerABC = None,
                 gradient_clipping: float = 1.0,
                 output_transform=None,
                 tensorboard_logs: str = None,
                 optional_tensorboard_features: bool=False,
                 embeddings_name: str = None,
                 finetune: bool = False):

        self.model: nn.Module = model

        self.forward_param_defaults = {}

        model_spec = inspect.getfullargspec(model.forward)
        self.forward_params: List[str] = model_spec.args[1:]
        for fparam, pdefault in zip(reversed(model_spec.args[1:]), reversed(model_spec.defaults if model_spec.defaults else [])):
            self.forward_param_defaults[fparam] = pdefault

        self.dataset_splits: DatasetSplits = dataset_splits
        self.loss: nn.Module = loss
        self.optimizer: optim.Optimizer = optimizer
        self.metrics: Dict[str, Metric] = metrics
        self.metrics: List[Metric] = [metric for _, metric in self.metrics.items()]
        self.experiment_config: ExperimentConfig = experiment_config
        self.device: str = device
        self.num_epochs: int = num_epochs
        self.scheduler: Any = scheduler
        self.seed: int = seed
        self.cuda: bool = cuda
        if self.cuda is None:  # If cuda not specified, just check if the cuda is available and use accordingly
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.loss_accumulation_steps: int = loss_accumulation_steps
        self.regularizer: RegularizerABC = regularizer
        self.gradient_clipping: float = gradient_clipping
        self.output_transform = output_transform
        self.tensorboard_logs: str = tensorboard_logs
        if self.tensorboard_logs:
            self.writer = SummaryWriter(log_dir=self.tensorboard_logs)
        self.optional_tensorboard_features: bool = optional_tensorboard_features
        self.embeddings_name = embeddings_name

        if self.output_transform:
            self.trainer, self.training_metrics = self.create_supervised_trainer(output_transform=self.output_transform)
            self.evaluator = self.create_supervised_evaluator(output_transform=self.output_transform)
        else:
            self.trainer, self.training_metrics = self.create_supervised_trainer()
            self.evaluator = self.create_supervised_evaluator()
        self.finetune = finetune

        self.optimizer_factory: PluginFactory = None

        loss_metrics = [m for m in self.metrics if isinstance(m, Loss)]

        if self.scheduler:
            if not loss_metrics:
                raise ValueError('A loss metric must be configured')
            elif len(loss_metrics) > 1:
                logging.warning('multiple loss metrics detected, using %s for LR scheduling', loss_metrics[0])
            self.loss_metric = loss_metrics[0]

        self.metrics_history = {"training": defaultdict(list),
                                "validation": defaultdict(list),
                                "test": defaultdict(list)}

        self.setup(self.training_metrics)

    def setup(self, training_metrics: Dict):
        def metric_name(n) -> str:
            if n.endswith('Accuracy'):
                n = 'acc'
            else:
                n = n[:-6] if n.endswith('Metric') else n
            return n

        def print_metrics(metrics) -> str:
            rv = ''
            metric_keys = sorted(k for k in metrics)
            for k in metric_keys:
                if k == 'Accuracy':
                    rv += f'{metric_name(k)}: {metrics[k]:.3}'
                else:
                    rv += f'{metric_name(k)}: {metrics[k]:.6}'
            return rv

        def store_metrics(metrics: Dict, mode: str):
            metric_keys = sorted(k for k in metrics)
            for k in metric_keys:
                self.metrics_history[mode][metric_name(k)].append(metrics[k])

        if self.seed:
            set_seed_everywhere(self.seed, self.cuda)

        pbar = ProgressBar()

        names = []
        for k, v in training_metrics.items():
            name = f'r{k}'
            names.append(name)
            RunningAverage(v).attach(self.trainer, name)
        RunningAverage(None, output_transform=lambda x: x[-1] * self.loss_accumulation_steps).attach(self.trainer, 'rloss')
        names.append('rloss')
        pbar.attach(self.trainer, names)

        pbar = ProgressBar()
        pbar.attach(self.evaluator)

        # A few events handler. To add / modify the events handler, you need to extend the __init__ method of RunnerABC
        # Ignite provides the necessary abstractions and a furnished repository of useful tools

        @self.trainer.on(Events.EPOCH_COMPLETED)
        def log_training_validation_results(trainer):

            self.evaluator.run(self.dataset_splits.train_data_loader())
            metrics = self.evaluator.state.metrics
            store_metrics(metrics=metrics, mode="training")
            logger.info(f"Training Results - Epoch: {trainer.state.epoch} {print_metrics(metrics)}")

            self.evaluator.run(self.dataset_splits.val_data_loader())
            metrics = self.evaluator.state.metrics
            store_metrics(metrics=metrics, mode="validation")
            logger.info(f"Validation Results - Epoch: {trainer.state.epoch} {print_metrics(metrics)}")

            if self.scheduler:
                self.scheduler.step(metrics[self.loss_metric.__class__.__name__])

        @self.trainer.on(Events.COMPLETED)
        def log_test_results(trainer):
            self.evaluator.run(self.dataset_splits.test_data_loader())
            metrics = self.evaluator.state.metrics
            store_metrics(metrics=metrics, mode="test")
            logger.info(f"Test Results - Epoch: {trainer.state.epoch} {print_metrics(metrics)}")

        if self.tensorboard_logs:
            tb_logger = TensorboardLogger(log_dir=self.tensorboard_logs)
            tb_logger.attach(self.trainer,
                             log_handler=OutputHandler(tag="training", output_transform=lambda loss: {
                                 'loss': loss}),
                             event_name=Events.ITERATION_COMPLETED)
            tb_logger.attach(self.evaluator,
                             log_handler=OutputHandler(tag="validation",
                                                       metric_names=["LossMetric"],
                                                       another_engine=self.trainer),
                             event_name=Events.EPOCH_COMPLETED)

            if self.optional_tensorboard_features:
                tb_logger.attach(self.trainer,
                                 log_handler=OptimizerParamsHandler(self.optimizer),
                                 event_name=Events.ITERATION_STARTED)
                tb_logger.attach(self.trainer,
                                 log_handler=WeightsScalarHandler(self.model),
                                 event_name=Events.ITERATION_COMPLETED)
                tb_logger.attach(self.trainer,
                                 log_handler=WeightsHistHandler(self.model),
                                 event_name=Events.EPOCH_COMPLETED)
                tb_logger.attach(self.trainer,
                                 log_handler=GradsScalarHandler(self.model),
                                 event_name=Events.ITERATION_COMPLETED)

            # This is important to close the tensorboard file logger
            @self.trainer.on(Events.COMPLETED)
            def end_tensorboard(trainer):
                logger.info("Training completed")
                tb_logger.close()

        if self.embeddings_name:
            @self.trainer.on(Events.COMPLETED)
            def log_embeddings(trainer):
                if hasattr(self.model, self.embeddings_name) and hasattr(self.dataset_splits, "vectorizer"):
                    logger.info(f"Logging embeddings ({self.embeddings_name}) to Tensorboard!")
                    embeddings = getattr(self.model, self.embeddings_name).weight.data
                    metadata = [str(self.dataset_splits.vectorizer.data_vocab._id2token[token_index]).encode('utf-8') for token_index in
                                range(embeddings.shape[0])]
                    self.writer.add_embedding(mat=embeddings, metadata=metadata, global_step=self.trainer.state.epoch)

    def _forward(self, batch):
        model_inputs = {}
        for p in self.forward_params:
            val = batch.get(p)
            if val is None:
                if p in self.forward_param_defaults:
                    val = self.forward_param_defaults[p]
                else:
                    raise ValueError(f'missing model parameter "{p}"')

            model_inputs[p] = val

        return self.model(**model_inputs)

    def create_supervised_trainer(self, prepare_batch=_prepare_batch, non_blocking=False,
                                  output_transform=lambda y_pred, y_target, loss: (y_pred, y_target, loss)):

        if self.device:
            self.model.to(self.device)

        # Gradient accumulation trick adapted from :
        # https://medium.com/huggingface/training-larger-batches-practical-tips-on-1-gpu-multi-gpu-distributed-setups-ec88c3e51255
        accumulation_steps = self.loss_accumulation_steps

        def _update(engine, batch):

            self.model.train()
            batch = prepare_batch(batch, device=self.device, non_blocking=non_blocking)
            y_pred = self._forward(batch)
            loss = self.loss(input=y_pred, target=batch['y_target'])

            # Add a regularisation term at train time only
            if self.regularizer:
                loss += self.regularizer.compute_penalty(model=self.model)

            loss /= accumulation_steps

            loss.backward()
            torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.gradient_clipping)

            if engine.state.iteration % accumulation_steps == 0:
                self.optimizer.step()
                self.optimizer.zero_grad()

            return output_transform(y_pred, batch['y_target'], loss.item())

        engine = Engine(_update)
        metrics = {}
        for i, metric in enumerate(self.metrics):
            if not isinstance(metric, Loss):
                n = metric.__class__.__name__
                tm = TrainingMetric(metric)
                metrics[n] = tm
                tm.attach(engine, n)

        return engine, metrics

    def create_supervised_evaluator(self, prepare_batch=_prepare_batch, non_blocking=False, output_transform=lambda y, y_pred: (y, y_pred)):

        if self.device:
            self.model.to(self.device)

        def _inference(engine, batch):
            self.model.eval()
            with torch.no_grad():
                batch = prepare_batch(batch, device=self.device, non_blocking=non_blocking)
                y_pred = self._forward(batch)
                return output_transform(y_pred, batch['y_target'])

        engine = Engine(_inference)

        for i, metric in enumerate(self.metrics):
            metric.attach(engine, f'{str(metric.__class__.__name__)}')

        return engine

    def freeze_and_replace_final_layer(self):
        """
        Freeze al layers and replace the last layer with a custom Linear projection on the predicted classes
        Note: this method assumes that the pre-trained model ends with a `classifier` layer, that we want to learn
        :return:
        """
        # freeze all layers
        for param in self.model.parameters():
            param.requires_grad = False

        # Number of input features to the final classification layer
        number_features = self.model.classifier.in_features

        # If `classifier` has several layers itself, this will only remove the last on, otherwise this does not contain anything
        features = list(self.model.classifier.children())[:-1]
        logger.info(f"Keeping layers {list(self.model.classifier.children())[:-1]} from the classifier layer")
        logger.info(f"Append layer {torch.nn.Linear(number_features, self.model.num_labels)} to the classifier")

        # Create the final linear layer for classification
        features.append(torch.nn.Linear(number_features, self.model.num_labels))
        self.model.classifier = torch.nn.Sequential(*features)
        self.model = self.model.to(self.device)

    def train(self):
        """
        Launch the ignite training pipeline
        If fine-tuning mode is granted in the config file, freeze all layers, replace classification layer by a Linear layer
        and reset the optimizer
        :return:
        """
        if self.finetune:

            logger.info(f"Fine-tuning the last classification layer to the data")
            trainer_key = [k for k, v in self.experiment_config.experiment.items() if v is self]
            if trainer_key:
                self.optimizer_factory = self.experiment_config.factories['optimizer']
            else:
                raise ValueError('this trainer object was not found in config')

            self.freeze_and_replace_final_layer()
            self.optimizer = self.optimizer_factory.create()

        self.trainer.run(self.dataset_splits.train_data_loader(), max_epochs=self.num_epochs)



from abc import ABC
from pathlib import Path

from transfer_nlp.plugins.config import ExperimentConfig


class ReporterABC(ABC):
    """
    Reporter implementations write reports about trained models. They should at least produce human readable reports,
    but can additionally produce reports that are easily machine-parsable.
    """

    def report(self, experiment_name: str, experiment: ExperimentConfig, report_dir: Path) -> float:
        """
        report the results of an experiment
        :param experiment_name: the name of the experiment.
        :param experiment: the completed experiment.
        :param report_dir: the directory in which to write the report
        :return: the key metric value, it's assumed higher is better.
        """

        pass


import logging
import re
from typing import List

logger = logging.getLogger(__name__)


class TokenizerABC:

    def __init__(self):
        pass

    def tokenize(self, text: str):
        raise NotImplementedError


class CustomTokenizer(TokenizerABC):

    def __init__(self):
        super().__init__()

    def tokenize(self, text: str) -> List[str]:
        """
        Basic text preprocessing
        :param text:
        :return:
        """

        text = text.lower()
        text = re.sub(r"([.,!?])", r" \1 ", text)
        text = re.sub(r"[^a-zA-Z.,!?]+", r" ", text)
        tokens = text.split(" ")
        if not tokens[-1]:
            tokens = tokens[:-1]

        return tokens


class CharacterTokenizer(TokenizerABC):

    def __init__(self):
        super().__init__()

    def tokenize(self, text: str) -> List[str]:
        """

        :param text:
        :return:
        """

        return [char for char in text.lower()]


if __name__ == "__main__":
    logging.info('')

    example = "Hello world!"
    tokenizer = CustomTokenizer()
    tokenized = tokenizer.tokenize(text=example)
    logger.info(f"{tokenized}")



import torch

from transfer_nlp.plugins.config import register_plugin


class RegularizerABC:

    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    def __str__(self):
        raise NotImplemented

    def compute_penalty(self, model: torch.nn.Module):
        raise NotImplementedError


@register_plugin
class L1(RegularizerABC):

    def __init__(self, alpha: float = 0.01) -> None:
        self.alpha = alpha

    def __call__(self, parameter: torch.Tensor) -> torch.Tensor:
        return self.alpha * torch.sum(torch.abs(parameter))

    def __str__(self):
        return f"L1(alpha={self.alpha})"

    def compute_penalty(self, model: torch.nn.Module):
        """
        Compute a penalty value uniformly over layers
        :param self:
        :param model:
        :return:
        """

        penalty = 0

        for name, parameter in model.named_parameters():
            penalty += self(parameter)

        return penalty


@register_plugin
class L2(RegularizerABC):

    def __init__(self, alpha: float = 0.01) -> None:
        self.alpha = alpha

    def __str__(self):
        return f"L2(alpha={self.alpha})"

    def __call__(self, parameter: torch.Tensor) -> torch.Tensor:
        return self.alpha * torch.sum(torch.pow(parameter, 2))

    def compute_penalty(self, model: torch.nn.Module):
        """
        Compute a penalty value uniformly over layers
        :param self:
        :param model:
        :return:
        """

        penalty = 0

        for name, parameter in model.named_parameters():
            penalty += self(parameter)

        return penalty



import logging

import torch


def describe(x: torch.Tensor):

    print("Type: {}".format(x.type()))
    print("Shape/size: {}".format(x.shape))
    print("Values: \n{}".format(x))


if __name__ == "__main__":

    tensor = torch.rand(size=(3, 4), dtype=torch.float64)
    describe(x=tensor)



import string
from collections import Counter
from typing import Dict, Tuple, List, Any

import numpy as np
import pandas as pd

from transfer_nlp.common.tokenizers import CustomTokenizer
from transfer_nlp.loaders.vocabulary import Vocabulary, SequenceVocabulary
from transfer_nlp.plugins.config import register_plugin


class Vectorizer:

    def __init__(self, data_file: str):
        self.data_file = data_file
        # self.df = pd.read_csv(data_file)

    def vectorize(self, input_string: str):
        raise NotImplementedError



import logging
from pathlib import Path
from typing import Dict, Tuple, Union

import numpy as np
import torch
from smart_open import open
from tqdm import tqdm

from transfer_nlp.loaders.loaders import DatasetSplits
from transfer_nlp.plugins.config import register_plugin
from transfer_nlp.plugins.helpers import ObjectHyperParams

logger = logging.getLogger(__name__)


def load_glove_from_file(glove_filepath: Path) -> Tuple[Dict[str, int], np.array]:
    w2i = {}
    embeddings = []

    with open(glove_filepath, "r") as fp:
        for index, line in tqdm(enumerate(fp), "Embeddings"):
            line = line.split(" ")  # each line: word num1 num2 ...
            w2i[line[0]] = index  # word = line[0]
            embedding_i = np.array([float(val) for val in line[1:]])
            embeddings.append(embedding_i)

    return w2i, np.stack(embeddings)


@register_plugin
class EmbeddingsHyperParams(ObjectHyperParams):

    def __init__(self, dataset_splits: DatasetSplits):
        super().__init__()
        self.words = dataset_splits.vectorizer.data_vocab._token2id.keys()


@register_plugin
class Embedding:

    def __init__(self, glove_filepath: Union[Path, str], data: DatasetSplits):

        words = data.vectorizer.data_vocab._token2id.keys()

        w2i, glove_embeddings = load_glove_from_file(glove_filepath=glove_filepath)
        embedding_size = glove_embeddings.shape[1]

        final_embeddings = np.zeros((len(words), embedding_size))

        for i, word in tqdm(enumerate(words), total=len(words), desc='Loading pre-trained word embeddings'):
            if word in w2i:
                final_embeddings[i, :] = glove_embeddings[w2i[word]]
            else:
                embedding_i = torch.ones(1, embedding_size)
                torch.nn.init.xavier_uniform_(embedding_i)
                final_embeddings[i, :] = embedding_i

        self.embeddings = final_embeddings



from typing import Dict, List, Tuple

import torch


def pretty_print(results: List[Tuple[str, torch.Tensor]]):
    """
    Pretty print embedding results.
    """
    for item in results:
        print("...[%.2f] - %s" % (item[1], item[0]))


def get_closest(target_word: str, word_to_idx: Dict, embeddings: torch.Tensor, n: int = 5) -> List[Tuple[str, torch.Tensor]]:
    """
    Get the n closest
    words to your word.
    """

    # Calculate distances to all other words

    word_embedding = embeddings[word_to_idx[target_word.lower()]]
    distances = []
    for word, index in word_to_idx.items():
        if word == "<MASK>" or word == target_word:
            continue
        distances.append((word, torch.dist(word_embedding, embeddings[index])))

    results = sorted(distances, key=lambda x: x[1])[1:n + 2]
    return results



"""
This file contains all necessary plugins classes that the framework will use to let a user interact with custom models, data loaders, etc...

The Registry pattern used here is inspired from this post: https://realpython.com/primer-on-python-decorators/
"""
import inspect
import json
import logging
import os
from abc import abstractmethod, ABC
from enum import Enum
from pathlib import Path
from typing import Dict, Union, Any, Optional, AbstractSet, Set, List

import ignite.metrics as metrics
import torch.nn as nn
import torch.optim as optim
from smart_open import open

logger = logging.getLogger(__name__)

CLASSES = {
    'CrossEntropyLoss': nn.CrossEntropyLoss,
    'BCEWithLogitsLoss': nn.BCEWithLogitsLoss,
    "Adam": optim.Adam,
    "SGD": optim.SGD,
    "AdaDelta": optim.Adadelta,
    "AdaGrad": optim.Adagrad,
    "SparseAdam": optim.SparseAdam,
    "AdaMax": optim.Adamax,
    "ASGD": optim.ASGD,
    "LBFGS": optim.LBFGS,
    "RMSPROP": optim.RMSprop,
    "Rprop": optim.Rprop,
    "ReduceLROnPlateau": optim.lr_scheduler.ReduceLROnPlateau,
    "MultiStepLR": optim.lr_scheduler.MultiStepLR,
    "ExponentialLR": optim.lr_scheduler.ExponentialLR,
    "CosineAnnealingLR": optim.lr_scheduler.CosineAnnealingLR,
    "LambdaLR": optim.lr_scheduler.LambdaLR,
    "ReLU": nn.functional.relu,
    "LeakyReLU": nn.functional.leaky_relu,
    "Tanh": nn.functional.tanh,
    "Softsign": nn.functional.softsign,
    "Softshrink": nn.functional.softshrink,
    "Softplus": nn.functional.softplus,
    "Sigmoid": nn.Sigmoid,
    "CELU": nn.CELU,
    "SELU": nn.functional.selu,
    "RReLU": nn.functional.rrelu,
    "ReLU6": nn.functional.relu6,
    "PReLU": nn.functional.prelu,
    "LogSigmoid": nn.functional.logsigmoid,
    "Hardtanh": nn.functional.hardtanh,
    "Hardshrink": nn.functional.hardshrink,
    "ELU": nn.functional.elu,
    "Softmin": nn.functional.softmin,
    "Softmax": nn.functional.softmax,
    "LogSoftmax": nn.functional.log_softmax,
    "GLU": nn.functional.glu,
    "TanhShrink": nn.functional.tanhshrink,
    "Accuracy": metrics.Accuracy,
}


def register_plugin(clazz):
    if clazz.__name__ in CLASSES:
        raise ValueError(f"{clazz.__name__} is already registered to class {CLASSES[clazz.__name__]}. Please select another name")
    else:
        CLASSES[clazz.__name__] = clazz
        return clazz


class UnknownPluginException(Exception):
    def __init__(self, clazz: str):
        super().__init__(f'Class {clazz} is not registered. See transfer_nlp.config.register_plugin for more information.')
        self.clazz: str = clazz


class UnconfiguredItemsException(Exception):
    def __init__(self, items: Dict[str, Set]):
        super().__init__(f'There are some unconfigured items, which makes these items not configurable: {items}')
        self.items: Dict[str, Set] = items


class BadParameter(Exception):
    def __init__(self, clazz, param):
        super().__init__(f"Parameter naming error: '{param}' is not a parameter of class '{clazz}'")
        self.param = param
        self.clazz = clazz


class ConfigFactoryABC(ABC):

    @abstractmethod
    def create(self):
        pass


class ParamFactory(ConfigFactoryABC):
    """
    Factory for simple parameters
    """

    def __init__(self, param):
        self.param = param

    def create(self):
        return self.param


class PluginFactory(ConfigFactoryABC):
    """
    Factory for complex objects creation
    """

    def __init__(self, cls, param2config_key: Optional[Dict[str, str]], *args, **kwargs):
        self.cls = cls
        self.param2config_key = param2config_key
        self.args = args
        self.kwargs = kwargs

    def create(self):
        return self.cls(*self.args, **self.kwargs)


class DefaultParamsMode(Enum):
    IGNORE_DEFAULTS = 0,
    NOT_IN_EXPERIMENT = 1,
    USE_DEFAULTS = 2


def _replace_env_variables(dico: Dict, env: Dict) -> None:
    """
    Replace all occurrences of environment variable to particular strings
    :param dico:
    :param env:
    :return:
    """
    env_keys = sorted(env.keys(), key=lambda k: len(k), reverse=True)

    def do_env_subs(v: Any) -> str:
        v_upd = v
        if isinstance(v_upd, str):
            for env_key in env_keys:
                env_val = env[env_key]

                if isinstance(env_val, os.PathLike):
                    env_val = str(env_val)

                if not isinstance(env_val, str):
                    if v == f'${env_key}':
                        # allow for non string value replacements
                        v_upd = env_val
                        break
                else:
                    v_upd = v_upd.replace('$' + env_key, env_val)

            if v_upd != v:
                logger.info('*** updating parameter %s -> %s', v, v_upd)

        return v_upd

    def recursive_replace(dico: Dict):

        for k, v in dico.items():
            if not isinstance(v, dict) and not isinstance(v, list):
                v = do_env_subs(v)
                dico[k] = v
            elif isinstance(v, list) and all(not isinstance(vv, dict) and not isinstance(vv, list) for vv in v):
                upd = []
                for vv in v:
                    upd.append(do_env_subs(vv))
                    dico[k] = upd
            elif isinstance(v, dict):
                recursive_replace(dico[k])
            else:
                pass

    recursive_replace(dico=dico)


class ExperimentConfig:

    @staticmethod
    def load_experiment_json(experiment: Union[str, Path, Dict]) -> Dict:
        if isinstance(experiment, dict):
            config = dict(experiment)
        else:
            with open(experiment) as f:
                config = json.load(f)
        return config

    def __init__(self, experiment: Union[str, Path, Dict], **env):
        """
        :param experiment: the experiment config
        :param env: substitution variables, e.g. a HOME directory. generally use all caps.
        :return: the experiment
        """
        self.factories: Dict[str, ConfigFactoryABC] = {}
        self.experiment: Dict[str, Any] = {}

        config = ExperimentConfig.load_experiment_json(experiment)
        _replace_env_variables(dico=config, env=env)

        # extract simple parameters
        logger.info(f"Initializing simple parameters:")
        for k, v in config.items():
            if not isinstance(v, dict) and not isinstance(v, list):
                logger.info(f"Parameter {k}: {v}")
                self.experiment[k] = v
                self.factories[k] = ParamFactory(v)

        # extract simple lists
        logger.info(f"Initializing simple lists:")
        for k, v in config.items():
            if isinstance(v, list) and all(not isinstance(vv, dict) and not isinstance(vv, list) for vv in v):
                logger.info(f"Parameter {k}: {v}")
                self.experiment[k] = v
                self.factories[k] = PluginFactory(list, None, v)

        for k in self.experiment:
            del config[k]

        self._build_items(config)

    def _do_recursive_build(self, object_key: str, object_dict: Dict, default_params_mode: DefaultParamsMode, unconfigured_keys: AbstractSet,
                            parent_level: str):

        def resolve_simple_value(factory_key: str, val: Any) -> Any:
            if isinstance(val, str):
                if val[0] == '$':
                    keyval = val[1:]
                    if keyval in self.experiment:
                        self.factories[factory_key] = self.factories[keyval]
                        return self.experiment[keyval]
                    else:
                        raise UnconfiguredItemsException({factory_key: {val}})

            return val

        def do_recursive_build_list(list_object: List, arg_name: str) -> List:
            """
            Function to recursively deal with nested lists
            :param list_object:
            :param arg_name:
            :return:
            """
            copy = []
            for i, element in enumerate(list_object):
                if isinstance(element, dict):
                    if "_name" in element:
                        element = self._do_recursive_build(object_key=str(i), object_dict=element,
                                                           default_params_mode=default_params_mode,
                                                           unconfigured_keys=unconfigured_keys,
                                                           parent_level=f'{parent_level}.{arg_name}.{i}')
                    else:
                        for key in element:
                            if isinstance(element[key], dict):
                                element[key] = self._do_recursive_build(object_key=key, object_dict=element[key],
                                                                       default_params_mode=default_params_mode,
                                                                       unconfigured_keys=unconfigured_keys,
                                                                       parent_level=f'{parent_level}.{arg}.{i}.{key}')

                            elif isinstance(element[key], list):
                                element[key] = do_recursive_build_list(list_object=element[key], arg_name=f"{arg_name}.{i}.{key}")
                            else:
                                element[key] = resolve_simple_value(f'{parent_level}.{arg}.{key}', element[key])
                        self.factories[f'{parent_level}.{arg}'] = PluginFactory(dict, None, list(element.items()))

                elif isinstance(element, list):
                    element = do_recursive_build_list(list_object=element, arg_name=f"{arg_name}.{i}")

                else:
                    element = resolve_simple_value(f'{parent_level}.{arg_name}.{i}', element)
                copy.append(element)

            result = copy
            self.factories[f'{parent_level}.{arg_name}'] = PluginFactory(list, None, list(copy))
            return result

        logger.info(f"Configuring {object_key}")

        if '_name' not in object_dict:
            raise ValueError(f"The object {object_key} should have a _name key to access its class")

        class_name = object_dict['_name']
        clazz = CLASSES.get(class_name)

        if not clazz:
            raise UnknownPluginException(object_dict["_name"])

        if inspect.isclass(clazz):
            spec = inspect.getfullargspec(clazz.__init__)
            spec_args = spec.args[1:]
        elif inspect.isfunction(clazz):
            spec = inspect.getfullargspec(clazz)
            spec_args = spec.args
        elif inspect.ismethod(clazz):
            spec = inspect.getfullargspec(clazz)
            spec_args = spec.args[1:]
        else:
            raise ValueError(f"{class_name} should be either a class, a function or a method")

        params = {}
        param2config_key = {}
        named_params = {p: pv for p, pv in object_dict.items() if p != '_name'}
        default_params = {p: pv for p, pv in zip(reversed(spec.args), reversed(spec.defaults))} if spec.defaults else {}

        for named_param in named_params:
            if named_param not in spec_args:
                raise BadParameter(clazz=class_name, param=named_param)

        for arg in spec_args:

            if arg == 'experiment_config':
                params[arg] = self
                param2config_key[arg] = arg

            elif arg in named_params:
                value = named_params[arg]

                if isinstance(value, dict):
                    if '_name' in value:
                        value = self._do_recursive_build(object_key=arg, object_dict=value,
                                                         default_params_mode=default_params_mode,
                                                         unconfigured_keys=unconfigured_keys,
                                                         parent_level=parent_level + "." + arg)
                    else:
                        for item in value:
                            if isinstance(value[item], dict):
                                value[item] = self._do_recursive_build(object_key=item, object_dict=value[item],
                                                                       default_params_mode=default_params_mode,
                                                                       unconfigured_keys=unconfigured_keys,
                                                                       parent_level=f'{parent_level}.{arg}.{item}')
                            else:  # value[item] is either an object defined in a dictionary, or it's an already built object
                                value[item] = resolve_simple_value(f'{parent_level}.{arg}.{item}', value[item])
                        self.factories[f'{parent_level}.{arg}'] = PluginFactory(dict, None, list(value.items()))

                elif isinstance(value, list):
                    value = do_recursive_build_list(list_object=value, arg_name=arg)

                elif isinstance(value, str) and value[0] == '$':

                    if value[1:] in self.experiment:
                        logger.info(f"Using the object {value}, already instantiated")
                        value = self.experiment[value[1:]]
                    else:
                        logger.info(f"{value} not configured yet, will be configured in next iteration")
                else:
                    logger.info(f"Using value {arg} / {named_params[arg]} from the config file")

                if not (isinstance(value, str) and value[0] == '$'):
                    params[arg] = value
                    param2config_key[arg] = value
                else:
                    logger.debug(f"You need to define '{value}' as a '{value[1:]}' object in the config file")

            # For values that are not in named_params, we look first at the experiment dict, then at the defaults parameters
            elif arg in self.experiment:
                params[arg] = self.experiment[arg]
                param2config_key[arg] = arg

            elif default_params_mode == DefaultParamsMode.NOT_IN_EXPERIMENT and \
                    arg not in self.experiment and arg not in unconfigured_keys and \
                    arg in default_params:
                params[arg] = default_params[arg]
                param2config_key[arg] = None
            elif default_params_mode == DefaultParamsMode.USE_DEFAULTS and arg in default_params:
                params[arg] = default_params[arg]
                param2config_key[arg] = None

        if len(params) == len(spec_args):
            self.factories[parent_level] = PluginFactory(cls=clazz, param2config_key=param2config_key, **params)
            return clazz(**params)

        else:
            unconfigured_params = spec_args - params.keys()
            raise UnconfiguredItemsException({parent_level: unconfigured_params})

    def _build_items_with_default_params_mode(self, config: Dict, default_params_mode: DefaultParamsMode):

        while config:
            configured = set()
            config_errors = {}
            for object_key, object_dict in config.items():

                try:
                    self.experiment[object_key] = self._do_recursive_build(object_key, object_dict,
                                                                           default_params_mode=default_params_mode,
                                                                           unconfigured_keys=config.keys(),
                                                                           parent_level=object_key)
                    configured.add(object_key)

                except BadParameter as b:
                    raise BadParameter(clazz=b.clazz, param=b.param)
                except UnconfiguredItemsException as e:
                    config_errors.update(e.items)
            if configured:
                for k in configured:
                    del config[k]

            else:
                if config_errors:
                    raise UnconfiguredItemsException(config_errors)

    def _build_items(self, config: Dict[str, Any]):

        try:
            logger.info(f"Initializing complex configurations ignoring default params:")
            self._build_items_with_default_params_mode(config, DefaultParamsMode.IGNORE_DEFAULTS)
        except UnconfiguredItemsException as e:
            pass

        try:
            logger.info(f"Initializing complex configurations only filling in default params not found in the experiment:")

            self._build_items_with_default_params_mode(config, DefaultParamsMode.NOT_IN_EXPERIMENT)
        except UnconfiguredItemsException as e:
            pass

        try:
            logger.info(f"Initializing complex configurations filling in all default params:")
            self._build_items_with_default_params_mode(config, DefaultParamsMode.USE_DEFAULTS)
        except UnconfiguredItemsException as e:
            logging.error('There are unconfigured items in the experiment. Please check your configuration:')
            for k, v in e.items.items():
                logging.error(f'"{k}" missing properties:')
                for vv in v:
                    logging.error(f'\t+ {vv}')

            raise e

    def _check_init(self):
        if self.experiment is None:
            raise ValueError('experiment config is not setup yet!')

    # map-like methods
    def __getitem__(self, item):
        self._check_init()
        return self.experiment[item]

    def get(self, item, default=None):
        self._check_init()
        return self.experiment.get(item, default)

    def __iter__(self):
        self._check_init()
        return iter(self.experiment)

    def items(self):
        self._check_init()
        return self.experiment.items()

    def values(self):
        self._check_init()
        return self.experiment.values()

    def keys(self):
        self._check_init()
        return self.experiment.keys()

    def __setitem__(self, key, value):
        raise ValueError("cannot update experiment!")



"""
This file contains an abstract CustomDataset class, on which we can build up custom dataset classes.

In your project, you will have to customize your data loader class. To let the framework interact with your class, you
need to use the decorator @register_dataset, just as in the examples in this file
"""

from torch.utils.data import Dataset, DataLoader

from transfer_nlp.loaders.vectorizers import Vectorizer
from transfer_nlp.plugins.config import register_plugin
from transfer_nlp.plugins.helpers import ObjectHyperParams


@register_plugin
class DatasetHyperParams(ObjectHyperParams):

    def __init__(self, vectorizer: Vectorizer):
        super().__init__()
        self.vectorizer = vectorizer


class DatasetSplits:
    def __init__(self,
                 train_set: Dataset, train_batch_size: int,
                 val_set: Dataset, val_batch_size: int,
                 test_set: Dataset = None, test_batch_size: int = None):
        self.train_set: Dataset = train_set
        self.train_batch_size: int = train_batch_size

        self.val_set: Dataset = val_set
        self.val_batch_size: int = val_batch_size

        self.test_set: Dataset = test_set
        self.test_batch_size: int = test_batch_size

    def train_data_loader(self):
        return DataLoader(self.train_set, self.train_batch_size, shuffle=True)

    def val_data_loader(self):
        return DataLoader(self.val_set, self.val_batch_size, shuffle=False)

    def test_data_loader(self):
        return DataLoader(self.test_set, self.test_batch_size, shuffle=False)


class DataFrameDataset(Dataset):

    def __init__(self, df):
        self.df = df

    def __len__(self):
        return len(self.df)

    def __getitem__(self, item):
        row = self.df.iloc[item, :]
        return {col: row[col] for col in self.df.columns}


class DataProps:
    def __init__(self):
        self.input_dims: int = None
        self.output_dims: int = None



import torch.nn as nn

from transfer_nlp.plugins.config import register_plugin


@register_plugin
class ObjectHyperParams:
    """
    Use or extend this class to configure model hyper parameters that cannot be predetermined. E.g.
    a model input size that depends on the data set composition.
    """

    def __init__(self):
        self.input_dim: int = None
        self.output_dim: int = None


@register_plugin
class TrainableParameters:
    """
    Use this class to configure optimizer parameters.
    """

    def __init__(self, model):
        self.model: nn.Module = model

    def __iter__(self):
        for p in self.model.parameters():
            if p.requires_grad:
                yield p



from setuptools import setup, find_packages
from io import open

setup(
    name='transfer_nlp',
    packages=find_packages(exclude=["*.tests", "*.tests.*",
                                    "tests.*", "tests"]),
    version='0.1.3',
    license='MIT',
    description='NLP library designed for flexible research and development',
    long_description=open("README.md", "r", encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    author='Peter Martigny',
    author_email='peter.martigny@gmail.com',
    url='https://github.com/feedly/transfer-nlp',
    download_url='https://github.com/feedly/transfer-nlp/archive/v0.1.3.tar.gz',
    keywords=['NLP', 'transfer learning', 'language models', 'NLU'],
    install_requires=[
        'tqdm==4.31.1',
        'matplotlib==3.0.3',
        'nltk==3.4',
        'pandas==0.24.1',
        'seaborn==0.9.0',
        'scipy==1.2.1',
        'annoy==1.15.1',
        'numpy==1.16.2',
        'requests==2.21.0',
        'torch==1.0.1.post2',
        'ipython==6.2.1',
        'feedly-client==0.20',
        'bs4==0.0.1',
        'tensorboardX==1.6',
        'knockknock==0.1',
        'pytorch-ignite==0.2.0',
        'smart_open==1.8.1'
    ],
    setup_requires=['green'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
    ],
)



from pathlib import Path
from typing import Dict, List

import numpy as np
from annoy import AnnoyIndex
from smart_open import open
from tqdm import tqdm


class PreTrainedEmbeddings(object):

    def __init__(self, word_to_index: Dict, word_vectors: List[np.array]):

        self.w2i: Dict = word_to_index
        self.word_vectors: List[np.array] = word_vectors
        self.i2w: Dict = {v: k for k, v in self.w2i.items()}

        self.index: AnnoyIndex = AnnoyIndex(len(word_vectors[0]), metric='euclidean')
        print("Building Index!")
        for _, i in tqdm(self.w2i.items(), 'Anoy Index'):
            self.index.add_item(i, self.word_vectors[i])
        self.index.build(50)
        print("Finished!")

    @classmethod
    def from_embeddings_file(cls, embedding_file: Path):

        w2i = {}
        word_vectors = []

        with open(embedding_file) as fp:
            for line in tqdm(fp.readlines(), 'Embeddings'):
                line = line.split(" ")
                word = line[0]
                vec = np.array([float(x) for x in line[1:]])

                w2i[word] = len(w2i)
                word_vectors.append(vec)

        return cls(word_to_index=w2i, word_vectors=word_vectors)

    def get_embedding(self, word: str):

        return self.word_vectors[self.w2i[word]]

    def get_closest_to_vector(self, vector: np.array, n: int = 1):

        nn_indices = self.index.get_nns_by_vector(vector, n)
        return [self.i2w[neighbor] for neighbor in nn_indices]

    def get_closest_to_word(self, word: str, n: int = 1):

        vector = self.get_embedding(word=word)
        return self.get_closest_to_vector(vector=vector, n=n)

    def compute_and_print_analogy(self, word1: str, word2: str, word3: str):

        vec1 = self.get_embedding(word=word1)
        vec2 = self.get_embedding(word=word2)
        vec3 = self.get_embedding(word=word3)

        spatial_relationship = vec2 - vec1
        vec4 = vec3 + spatial_relationship

        closest_words = self.get_closest_to_vector(vector=vec4, n=4)
        existing_words = set([word1, word2, word3])
        closest_words = [word for word in closest_words
                         if word not in existing_words]

        if len(closest_words) == 0:
            print("Could not find nearest neighbors for the computed vector!")
            return

        for word4 in closest_words:
            print("{} : {} :: {} : {}".format(word1, word2, word3, word4))


if __name__ == "__main__":
    embedding_file = Path.home() / 'work/experiments/nlp/data/glove/glove.6B.100d.txt'
    embeddings = PreTrainedEmbeddings.from_embeddings_file(embedding_file=embedding_file)

    print(embeddings.get_closest_to_word(word='house', n=10))
    embeddings.compute_and_print_analogy('man', 'he', 'woman')



from collections import Counter
from pathlib import Path
from typing import List, Dict
from tqdm import tqdm

import numpy as np
import pandas as pd
from nltk.tokenize import word_tokenize


class NgramLM:

    def __init__(self, n: int):

        self.n = n  # Number of tokens in N-gram windows
        self.data: pd.DataFrame = None
        self.epsilon = 1e-8  # proba for N-grams that were not seen in the corpus
        self.buckets: Dict = None  # will contain counts for every n-grams, n from 1 to self.n
        self.cutoff = 10

    def load_data(self, data_path: Path):
        """

        :param data_path:
        :return:
        """

        self.data = pd.read_csv(filepath_or_buffer=data_path)

    def tokenize(self):
        """

        :return:
        """

        self.data['content'] = self.data['content'].apply(lambda x: str(x).lower())
        self.data['content'] = self.data['content'].apply(word_tokenize)

    def ngramize(self):
        """
        Compute the counts for each bucket of ngrams
        :return:
        """

        data = [item for sublist in self.data.content for item in sublist]
        buckets = {i: [' '.join(data[j:j+i]) for j in range(0, len(data), i)] for i in tqdm(range(1, self.n), f"N-gram bucket")}
        self.buckets = {key: Counter(value) for key, value in buckets.items()}
        self.buckets = {key: Counter(el for el in value.elements() if value[el] >= self.cutoff) for key, value in self.buckets.items()}
        # for key in self.buckets:
        #     self.buckets[i] =
        # self.buckets = {key: {word: value[word] for word in value if int(value[word]) > self.cutoff} for key, value in buckets.items()}

    def proba(self, n_plus_one: List[str], n: List[str]) -> float:
        """
        Compute the probability of next word given the past:
        P(next word | window) = P(window + word) / P(window)
        If window + word does not exist in corpus, we return self.epsilon
        If window does not exist in corpus, w decrease size of window until it exists
        :param n_plus_one:
        :param n:
        :return:
        """

        size = len(n_plus_one)
        n_plus_one = ' '.join(n_plus_one)
        if n_plus_one not in self.buckets[size]:
            # print(f"using default small value since {n_plus_one} is not in the vocabulary")
            return self.epsilon

        numerator = self.buckets[size][n_plus_one]

        size = len(n)
        while ' '.join(n) not in self.buckets[size] and len(n) > 1:
            size -= 1
            n = n[1:]

        n = ' '.join(n)
        denominator = self.buckets[size][n]

        return numerator / denominator

    def build_from_dataframe(self, data_path: Path):
        """
        Build the LM from a dataframe path
        :param data_path:
        :return:
        """

        self.load_data(data_path=data_path)
        self.tokenize()
        self.ngramize()

    def generate(self, first_token: str, number_token: int) -> str:
        """
        Generate text based on a first token
        :param first_token:
        :param number_token:
        :return:
        """

        if first_token not in self.buckets[1]:
            return f"{first_token} does not belong to our Vocabulary"

        result = [first_token]
        for i in range(number_token):
            probas = {word: self.proba(n_plus_one=result[-self.n + 2:] + [word], n=result[-self.n + 2:]) for word in self.buckets[1]}

            # Sample from the probability distribution over next word
            p = list(probas.values())
            p /= np.sum(p)
            next_word = np.random.choice(a=list(probas.keys()), size=1, p=p)[0]
            result.append(next_word)

        return ' '.join(result)


if __name__ == '__main__':

    save_path = Path.home() / 'work/experiments/nlp/data/feedly_data10000.csv'
    lm = NgramLM(n=5)
    lm.build_from_dataframe(data_path=save_path)
    lm.generate(first_token='the', number_token=20)



import inspect
import logging
from itertools import zip_longest
from typing import Dict, List, Any

import torch
from ignite.utils import convert_tensor

from transfer_nlp.loaders.vectorizers import Vectorizer

logger = logging.getLogger(__name__)


def _prepare_batch(batch: Dict, device=None, non_blocking: bool = False):
    """Prepare batch for training: pass to a device with options.

    """
    result = {key: convert_tensor(value, device=device, non_blocking=non_blocking) for key, value in batch.items()}
    return result


class PredictorABC:

    def __init__(self, vectorizer: Vectorizer, model: torch.nn.Module):

        self.model: torch.nn.Module = model
        self.model.eval()
        self.forward_params = {}
        model_spec = inspect.getfullargspec(self.model.forward)
        for fparam, pdefault in zip_longest(reversed(model_spec.args[1:]), reversed(model_spec.defaults if model_spec.defaults else [])):
            self.forward_params[fparam] = pdefault

        self.vectorizer: Vectorizer = vectorizer

    def forward(self, batch: Dict[str, Any]) -> torch.tensor:
        """
        Do the forward pass
        :param batch:
        :return:
        """
        with torch.no_grad():
            batch = _prepare_batch(batch, device="cpu", non_blocking=False)
            model_inputs = {}
            for p, pdefault in self.forward_params.items():
                val = batch.get(p)
                if val is None:
                    if pdefault is None:
                        raise ValueError(f'missing model parameter "{p}"')
                    else:
                        val = pdefault

                model_inputs[p] = val
            y_pred = self.model(**model_inputs)

        return y_pred

    def json_to_data(self, input_json: Dict) -> Dict:
        """
        Transform a json entry into a data example, which is the same that what the __getitem__ method in the
        data loader, except that this does not output any expected label as in supervised setting
        :param input_json:
        :return:
        """
        raise NotImplementedError

    def output_to_json(self, *args, **kwargs) -> Dict[str, Any]:
        """
        Convert the result into a proper json
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def decode(self, *args, **kwargs) -> List[Dict]:
        """
        Return an output dictionary for every example in the batch
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def predict(self, batch: Dict[str, Any]) -> List[Dict]:
        """
        Decode the output of the forward pass
        :param batch:
        :return:
        """
        forward = self.forward(batch=batch)
        return self.decode(forward)

    def json_to_json(self, input_json: Dict) -> Dict[str, Any]:
        """
        Full prediction: input_json --> data example --> predictions --> json output
        :param input_json:
        :return:
        """
        json2data = self.json_to_data(input_json=input_json)
        predictions = self.predict(batch=json2data)
        predictions2json = self.output_to_json(predictions)

        return predictions2json
